<?php

namespace Drupal\billwerk_custom_example;

use Drupal\billwerk\BillwerkSync;
use Drupal\Core\Entity\ContentEntityBase;
use Drupal\scheduled_executable\Entity\ScheduledExecutable;
use Drupal\user\Entity\User;

/**
 * Class BillwerkCustomSync.
 *
 * Provides an alternative project specific class for the subman.sync service
 * class BillwerkSync of the billwerk module.
 *
 * @package Drupal\billwerk_custom_example
 */
class BillwerkCustomSync extends BillwerkSync {

  /**
   * {@inheritdoc}
   */
  public function updateSubscriberCustomSteps(array $data, ContentEntityBase $subscriber) {
    if ($subscriber->hasField('field_number_of_max_subaccounts')) {
      $max_subaccounts = 0;

      // Iterate through all valid subscriptions of the subscriber.
      // Determine the maximum number of subaccounts by the highest number found
      // via the plan variant's InternalName if it starts with "FA" or "SAH".
      foreach ($this->getSubscriptionsOfSubscriber($subscriber) as $subscription) {
        $subscription_variant_id = $subscription[self::MAP_SUBSCRIBER_NORM][self::MAP_SUBSCRIPTION_VARIANT];
        $subscription_variant = $this->getSubscriptionType($subscription_variant_id);
        if ($subscription_variant) {
          [$plan_name_alpha, $plan_name_number] = sscanf($subscription_variant['InternalName'], "%[A-Z]%d");
          if (in_array($plan_name_alpha, ['FA', 'SAH'])) {
            $max_subaccounts = max($max_subaccounts, $plan_name_number);
          }
        }
      }

      // Set determined number of subaccounts in respective field.
      $subscriber->set('field_number_of_max_subaccounts', $max_subaccounts);
    }

    // For Drupal users newly created by the Billwerk interface...
    if ($subscriber->isNew()) {
      // Assign email address vom Billwerk customer only once when Drupal user is newly created,
      // to allow them to stay independent afterwards.
      // (Default sync of email address is deactivated by removing email from normalized data, via overridden getNormalizedData())
      $subscriber->set('mail', $data[self::DATA_KEY_SUBSCRIBER]['EmailAddress']);

      // Assign "Kundenummer" from Billwerk.
      $subscriber->set('field_customer_number', $data[self::DATA_KEY_SUBSCRIBER][self::MAP_SUBSCRIBER_NORM][self::MAP_SUBSCRIBER_FOREIGN_ID]);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getNormalizedData($type, $data) {
    // Let parent class do default normalization.
    $norm = parent::getNormalizedData($type, $data);

    // Remove email address from normalized data,
    // to allow them to stay independent afterwards.
    // (First assignment for newly created users is done in overridden updateSubscriberCustomSteps())
    if ($type == self::DATA_KEY_SUBSCRIBER) {
      unset($norm[self::MAP_SUBSCRIBER_EMAIL]);
    }

    return $norm;
  }

  /**
   * {@inheritdoc}
   */
  public function performAdditionalWebhookAction($event, $user) {
    switch ($event['event_name']) {
      case 'ContractCreated':
        // Schedule a TrialEndApproaching notification being sent to the user:
        // Create an instance of our subman_notify_trial_end_approaching action.
        $actionManager = \Drupal::service('plugin.manager.action');
        $action = $actionManager->createInstance('subman_notify_trial_end_approaching');

        $trial_end_notification_days = $this->utils->config('trial_end_notification_days', FALSE);

        if ($trial_end_notification_days && $user) {
          // Get trial end date of user.
          $trial_end_time = $this->getTrialEndForUser($user);

          // If a trial end was found, create an ScheduledExecutable.
          if ($trial_end_time) {
            $execution_time = max(time(), $trial_end_time - $trial_end_notification_days * 24 * 60 * 60);

            // Create a scheduled_executable with the action and the user.
            $scheduled_executable = ScheduledExecutable::create()
              ->setExecutablePlugin('action', $action)
              ->setTargetEntity($user)
              ->setResolver('default')
              ->setKey('trial_end_approaching:' . $user->id())
              ->setExecutionTime($execution_time);
            $scheduled_executable->save();

            // Log the creation of the scheduled executable.
            $this->utils->log('Scheduled trial end mail for subscriber @entity_id on @trial_end_time.', NULL, [
              '@entity_id' => $user->id(),
              '@trial_end_time' => date('Y-m-d H:i:s', $trial_end_time),
            ]);
          }
        }
        break;
    }
  }

  /**
   * Retrieves if/when any trial period ends for the given user.
   *
   * @param \Drupal\user\Entity\User $user
   *   The user to check for.
   *
   * @return false|int The unix timestamp of the trial end, or FALSE if the user is not in an active uncancelled trial.
   */
  public function getTrialEndForUser(User $user) {
    $trial_end = FALSE;

    // Only proceed if the user has the role subscriber_trial but _not_ the role license_educational.
    $trial_role = $user->hasRole('subscriber_trial') && !$user->hasRole('license_educational');
    if (!$trial_role) {
      return $trial_end;
    }

    // Check subscriptions of the user for a trial subscription (SPA01).
    $subscriptions = $this->getSubscriptionsOfSubscriber($user, TRUE);
    foreach ($subscriptions as $subscription) {
      $norm = $subscription[self::MAP_SUBSCRIBER_NORM];

      // Determine whether the subscription is a trial subscription (by its variant's name = SPA01).
      $this->getSubscriptionTypes(TRUE);
      $subscription_type = $this->getSubscriptionType($norm[self::MAP_SUBSCRIPTION_VARIANT]);
      $subscription_type_name_trial = $this->utils->config('subscription_name_trial', 'SPA01');
      $subscription_type_name_technical = $subscription_type[self::MAP_SUBSCRIBER_NORM][self::MAP_SUBSCRIPTION_TYPE_NAME_TECHNICAL];
      $is_trial_subscription = ($subscription_type_name_technical == $subscription_type_name_trial);

      // If we have a trial subscription, check whether it's active=uncancelled).
      if ($is_trial_subscription && $subscription['LifecycleStatus'] == 'Active') {
        // Get trial end date from the start date of a second pahse, if there is
        // a second contrac phase defined (which would be a regular plan).
        $phase2_type = $subscription['Phases'][1]['Type'];
        $phase2_start_date = $subscription['Phases'][1]['StartDate'] ?? FALSE;
        if (($phase2_type == 'Normal') && $phase2_start_date) {
          $trial_end = strtotime($phase2_start_date);
        }
      }
    }

    return $trial_end;
  }

  /**
   * Checks whether the given user currently only has a non-terminated trial.
   *
   * @param $user
   *   User The user to check for.
   */
  public function notifyUserOfTrialEnd($user) {
    // Check whether user is active and has still got a trial subscription with not contract end date yet (i.e. not yet terminated).
    $sync_data = $this->getSyncDataOfUser($user);
    if ($user->isActive()) {
      // Get trial end date of user - if it returns FALSE, the user is not in an active uncanceled trial anymore.
      $trial_end_time = $this->getTrialEndForUser($user);

      // If a trial end was found, send the mail.
      if ($trial_end_time) {
        // Notify the user via email.
        $this->notifyUser($user, self::NOTIFY_TRIAL_ENDING);

        // Log the sending of the email.
        $this->utils->log('Send trial end mail notification to subscriber @entity_id.', NULL, [
          '@entity_id' => $user->id(),
        ]);
      }
    }
  }

}
