<?php

namespace Drupal\billwerk;

use Drupal\subman\Event\SubmanIncomingWebhook;
use Drupal\subman\Event\SubmanSynchronousResult;
use Drupal\subman\SubmanSync;
use Drupal\subman\SubmanUtilities;
use Drupal\subman\SubmanWebhookException;
use Drupal\user\Entity\User;

/**
 * Class BillwerkSync.
 *
 * This is where the generic SubmanSync class is extended for interfacing with the specific subscription management SaaS Billwerk, by implementing some abstract methods and overriding/extending other methods.
 *
 * As a specific implementation of the abstrac SubmanSync class, this class provides:
 * * handling of the webhook event
 * * delegating remote api data lookups to class BillwerkApi
 * * normalizing SaaS specific data to a normalized format that
 *   the methods of generic class SubmanSync understands
 *
 * This class is declared as service "subman.sync", and as an event subscriber,
 * using the service tag "event_subscriber".
 * We assume that only one implementation of the abstract class SubmanSync
 * is used at a time. This is an easy way to let other code use a specific
 * service class for a specific SaaS without knowing about it specifically.
 *
 * Any project specific customizations can be done by overriding the
 * "subman.sync" service with a custom class extending this class.
 */
class BillwerkSync extends SubmanSync implements BillwerkSyncInterface {

  /**
   * Drupal\billwerk\BillwerkApi definition.
   *
   * @var \Drupal\billwerk\BillwerkApi
   */
  protected $api;

  /**
   * Constructs a new BillwerkSync object.
   */
  public function __construct(SubmanUtilities $subman_utilities, BillwerkApi $billwerk_api) {
    parent::__construct($subman_utilities);
    $this->api = $billwerk_api;
  }

  public function getSubscriptionManagementServiceTitle() {
    return 'Billwerk';
  }

  /**
   * Handle an incoming webhook event.
   *
   * @param \Drupal\subman\Event\SubmanIncomingWebhook $event
   *   The event with the data webhook data.
   *
   * @throws \Drupal\subman\SubmanWebhookException
   */
  public function onIncomingWebhook(SubmanIncomingWebhook $event) {
    $json = $event->dataJson;
    SubmanWebhookException::throwOnEmpty($json, 'Empty or erroneous JSON data.');

    // Initialize basic variables from data.
    $event = [
      'event_name' => $json['Event'],
      'customer_id' => $json['CustomerId'] ?? '',
      'contract_id' => $json['ContractId'] ?? '',
    ];

    // Log event.
    $this->utils->log($event['event_name'] . ' for customer_id:' . $event['customer_id'] . '/contract_id:' . $event['contract_id'], $json);

    // Process event.
    return $this->processIncomingWebhook($event);
  }

  /**
   * @inheritdoc
   */
  public function onSynchronousResult(SubmanSynchronousResult $event) {
    $id = $event->request->get('contractId');
    SubmanWebhookException::throwOnEmpty($id, 'Empty contractId.');

    $subscription = $this->getSubscriptionData($id);
    $subscription_type = $this->getSubscriptionType($subscription[self::MAP_SUBSCRIBER_NORM][self::MAP_SUBSCRIPTION_TYPE]);
    $subscription_variant = $this->getSubscriptionType($subscription[self::MAP_SUBSCRIBER_NORM][self::MAP_SUBSCRIPTION_VARIANT]);

    $this->updateSubscriberByExternalId($subscription[self::MAP_SUBSCRIBER_NORM][self::MAP_SUBSCRIPTION_SUBSCRIBER_ID]);
    $event->redirectUrl .= '?'
      .  'subscription_type_id='      . $subscription_type[self::MAP_SUBSCRIBER_NORM][self::MAP_SUBSCRIPTION_TYPE_ID]
      . '&subscription_type_name='    . $subscription_type[self::MAP_SUBSCRIBER_NORM][self::MAP_SUBSCRIPTION_TYPE_NAME]
      . '&subscription_variant_id='   . $subscription_variant[self::MAP_SUBSCRIBER_NORM][self::MAP_SUBSCRIPTION_TYPE_ID]
      . '&subscription_variant_name=' . $subscription_variant[self::MAP_SUBSCRIBER_NORM][self::MAP_SUBSCRIPTION_TYPE_NAME]
      ;
  }

  /**
   * Process an incoming webhook event.
   *
   * For simplicity sake, we always react on events with a full user update,
   * i.e. we always retrieve the subscriber and its subscriptions from the SaaS
   * and store everything in the respective Drupal user.
   *
   * @param array $event
   *   The structured event data, including relevant gathered ids in key '_IDs'.
   *
   * @throws \Drupal\subman\SubmanWebhookException
   */
  public function processIncomingWebhook($event) {
    // Act on specific events arriving from the subscription management SaaS:
    switch ($event['event_name']) {
      case 'Test':
        break;
      // customer_id based events:
      case 'CustomerChanged':
      case 'CustomerCreated':
      case 'CustomerDeleted':
      case 'CustomerLocked':
      case 'CustomerUnlocked':
      case 'PaymentEscalated':
      case 'PaymentEscalationReset':
      // Other potential events that contain as customer_id:
      // case 'DebitAuthCancelled':
        // Update subscriber.
        $this->updateSubscriberByExternalId($event['customer_id']);
        // Prepare any (custom) action.
        $user = $this->getSubscriberById($event['customer_id']);
        $this->performAdditionalWebhookAction($event, $user);
        break;
      // contract_id based events:
      case 'ContractCreated':
      case 'ContractChanged':
      case 'ContractCancelled':
      // Other potential events that contain only a contract_id:
      // case 'PaymentBearerExpiring':
      // case 'PaymentBearerExpired':
      // case 'PaymentDataChanged':
      // case 'PaymentProcessStatusChanged':
      // case 'PaymentRegistered':
        // Get subscription data and subscriber id.
        $subscription = $this->getSubscriptionData($event['contract_id']);
        $subscriber_id = $subscription[self::MAP_SUBSCRIBER_NORM][self::MAP_SUBSCRIPTION_SUBSCRIBER_ID];
        // Update subscriber.
        $this->updateSubscriberByExternalId($subscriber_id);
        // Prepare any (custom) action.
        $user = $this->getSubscriberById($subscriber_id);
        $this->performAdditionalWebhookAction($event, $user);
        break;
      // Notifications by contract_id.
      case 'TrialEndApproaching':
        // @Todo break out this case into a dedicated method.
        $subscription = $this->getSubscriptionData($event['contract_id']);
        $subscriber_data = $this->getSubscriberData($subscription[self::MAP_SUBSCRIBER_NORM][self::MAP_SUBSCRIPTION_SUBSCRIBER_ID]);
        $user = $this->lookupSubscriber($subscriber_data, FALSE);
        $this->notifyUser($user, self::NOTIFY_TRIAL_ENDING);
        break;
      // If we don't know how to handle the incoming event, raise an exception.
      default:
        throw new SubmanWebhookException(SubmanWebhookException::unsupportedEvent, $event['event_name']);
    }
  }

  /**
   * @inheritdoc
   */
  public function getNormalizedData($type, $data) {
    $normalized = parent::getNormalizedData($type, $data);

    if (!empty($data)) {
      if ($type == self::DATA_KEY_SUBSCRIBER) {
        $deleted = !empty($data['DeletedAt']);
        $normalized = array_merge([
          self::MAP_SUBSCRIBER_ID => $data['Id'],
          self::MAP_SUBSCRIBER_EMAIL => $data['EmailAddress'],
          self::MAP_SUBSCRIBER_FIRSTNAME => $data['FirstName'],
          self::MAP_SUBSCRIBER_LASTNAME => $data['LastName'],
          self::MAP_SUBSCRIBER_FOREIGN_ID => $data['DebitorAccount'] ?? '',
          self::MAP_SUBSCRIBER_CREATED => strtotime($data['CreatedAt']),
          self::MAP_SUBSCRIBER_ACTIVE => !$data['IsLocked'] && !$deleted,
          self::MAP_SUBSCRIBER_DELETED => $deleted,
        ], $normalized);
      }

      if ($type == self::DATA_KEY_SUBSCRIPTION) {
        // @Todo Clarify sensible default for dunning level.
        $escalation_threshold = $this->utils->config('subscription.escalation_revoke_threshold', FALSE);
        $valid_phases = $this->utils->config('subscription.valid_phases', ['Trial', 'Active', 'Normal']);
        $escalation = isset($data['CurrentDunning']) ? $data['CurrentDunning']['Level'] : FALSE;
        $escalation_invalid = ($escalation_threshold === FALSE || $escalation === FALSE) ? FALSE : ($escalation >= $escalation_threshold);
        $valid = !$escalation_invalid && in_array($data['CurrentPhase']['Type'], $valid_phases);
        $normalized = array_merge([
          self::MAP_SUBSCRIPTION_ID => $data['Id'],
          self::MAP_SUBSCRIPTION_SUBSCRIBER_ID => $data['CustomerId'],
          self::MAP_SUBSCRIPTION_TYPE => $data['PlanId'] ?? NULL,
          self::MAP_SUBSCRIPTION_VARIANT => $data['PlanVariantId'] ?? NULL,
          self::MAP_SUBSCRIPTION_REFERENCE => $data['ReferenceCode'],
          self::MAP_SUBSCRIPTION_MESSAGE => '',
          self::MAP_SUBSCRIPTION_ESCALATION => $escalation,
          self::MAP_SUBSCRIPTION_VALID => $valid,
        ], $normalized);
      }

      if ($type == self::DATA_KEY_SUBSCRIPTION_TYPE) {
        $is_plan = isset($data['Name']['_c']);
        $plan_name = $data['Name']['_c'] ?? NULL;
        $description = $data['PlanDescription']['_c'] ?? '';
        $internal_name = $data['InternalName'] ?? '';
        $variant_name = !$is_plan ? $description  . ' (' . $internal_name . ')' : NULL; // Used for cosmetic output.
        $variant_name_technical = !$is_plan ? $internal_name : NULL; // Used for technical decisionmaking.
        $normalized = array_merge([
          // Determine if this is a Plan or PlanVariant:
          self::MAP_SUBSCRIPTION_TYPE_CLASS => $is_plan ? self::MAP_SUBSCRIPTION_TYPE_CLASS_MAIN : self::MAP_SUBSCRIPTION_TYPE_CLASS_VARIANT,
          self::MAP_SUBSCRIPTION_TYPE_ID => $data['Id'],
          // Get Name for Plan or PlanVariant:
          self::MAP_SUBSCRIPTION_TYPE_NAME => $plan_name ?? $variant_name ?? NULL,
          self::MAP_SUBSCRIPTION_TYPE_NAME_TECHNICAL => $plan_name ?? $variant_name_technical ?? NULL,
          self::MAP_SUBSCRIPTION_TYPE_ROLES => $this->utils->explodeValues($data['CustomFields']['cms_roles'] ?? ''),
        ], $normalized);
      }
    }

    return $normalized;
  }

  /**
   * @inheritdoc
   */
  public function retrieveSubscriberDataViaExternalKey($key, $key_is_primary_id = TRUE) {
    if ($key_is_primary_id) {
      return $this->api->getCustomer($key);
    }
    else {
      // @Todo as soon as Billwerk offers a search for specific fields, try to use that.
      //       (the current search across multiple fields might turn up false positives)
      $results = $this->api->searchCustomers($key, 1);
      return empty($results) ? NULL: reset($results);
    }
  }

  /**
   * @inheritdoc
   */
  public function retrieveSubscriberData($external_id) {
    return $this->api->getCustomer($external_id);
  }

  /**
   * @inheritdoc
   */
  public function retrieveSubscriptionData($external_id) {
    return $this->api->getContract($external_id);
  }

  /**
   * @inheritdoc
   */
  public function retrieveSubscriptionsData($external_id) {
    return $this->api->getContractsForCustomer($external_id);
  }

  /**
   * @inheritdoc
   */
  public function retrieveSubscriptionTypesData() {
    return array_merge($this->api->getPlans(), $this->api->getPlanVariants());
  }

  /**
   * @inheritdoc
   */
  public function buildEmbedSignup($embed_parameter, $mail = FALSE) {
    $url = $this->api->getUrlSignup($embed_parameter);
    $url .= $mail ? '?email=' . urlencode($mail) : '';
    return $this->buildEmbed($url);
  }

  /**
   * @inheritdoc
   */
  public function buildEmbedSelfservice($embed_parameter) {
    // @Todo Add security check so that only a subscription that actually belongs to a subscriber is embedded.
    $url = $this->api->getUrlSelfservice($embed_parameter);
    return $this->buildEmbed($url);
  }

}
